@@ -1,4 +1,4 @@
-# Copyright (c) 2004-2020 mudpy authors. Permission to use, copy,
+# Copyright (c) 2004-2021 mudpy authors. Permission to use, copy,
 # modify, and distribute this software is granted under terms
 # provided in the LICENSE file distributed with this software.
 
@@ -11,6 +11,8 @@ import sys
 import telnetlib
 import time
 
+import yaml
+
 pidfile = "var/mudpy.pid"
 
 test_account0_setup = (
@@ -331,17 +333,32 @@ test_show_element = (
         r'  \x1b\[32mgender: \x1b\[31mfemale.*> ', ""),
 )
 
-test_show_result = (
-    (2, "> ", "show result 12345*67890"),
-    (2, r"\r\n838102050\r\n.*> ", "show result 1/0"),
-    (2, r"Your expression raised an exception.*division by zero.*> ",
-     "show result mudpy"),
-    (2, r"<module 'mudpy' from .*> ", "show result re"),
-    (2, r"Your expression raised an exception.*name 're' is not defined.*> ",
-     "show result universe"),
-    (2, r"<mudpy\.misc\.Universe object at 0x.*> ", "show result actor"),
-    (2, r"Your expression raised an exception.*name 'actor' is not "
-        r"defined.*> ", ""),
+test_evaluate = (
+    (2, "> ", "evaluate 12345*67890"),
+    (2, r"\r\n838102050\r\n.*> ", "evaluate 1/0"),
+    (2, "Your expression raised an exception.*division by zero.*> ",
+     "evaluate mudpy"),
+    (2, "<module 'mudpy' from.*> ", "evaluate re"),
+    (2, "Your expression raised an exception.*name 're' is not defined.*> ",
+     "evaluate universe"),
+    (2, r"<mudpy\.misc\.Universe object at 0x.*> ", "evaluate actor"),
+    (2, "Your expression raised an exception.*name 'actor' is not defined.*> ",
+        "evaluate dir(mudpy)"),
+    (2, "__builtins__.*> ", "evaluate mudpy.__builtins__.open"),
+    (2, "not allowed.*> ", "evaluate (lambda x: x + 1)(2)"),
+    (2, "not allowed.*> ", ""),
+)
+
+test_debug_restricted = (
+    (0, "> ", "help evaluate"),
+    (0, r"That is not an available command\.", "evaluate"),
+    (0, '(not sure what "evaluate" means|Arglebargle, glop-glyf)', ""),
+)
+
+test_debug_disabled = (
+    (2, "> ", "help evaluate"),
+    (2, r"That is not an available command\.", "evaluate"),
+    (2, '(not sure what "evaluate" means|Arglebargle, glop-glyf)', ""),
 )
 
 test_show_log = (
@@ -378,63 +395,72 @@ final_cleanup = (
     (2, r"Disconnecting\.\.\.", ""),
 )
 
-dialogue = (
-    (test_account0_setup, "first account setup"),
-    (test_account1_setup, "second account setup"),
-    (test_actor_appears, "actor spontaneous appearance"),
-    (test_explicit_punctuation, "explicit punctuation"),
-    (test_implicit_punctuation, "implicit punctuation"),
-    (test_typo_replacement, "typo replacement"),
-    (test_sentence_capitalization, "sentence capitalization"),
-    (test_chat_mode, "chat mode"),
-    (test_wrapping, "wrapping"),
-    (test_forbid_ansi_input, "raw escape input is filtered"),
-    (test_escape_macros, "replacement macros are escaped"),
-    (test_movement, "movement"),
-    (test_actor_disappears, "actor spontaneous disappearance"),
-    (test_abort_avatar_deletion, "abort avatar deletion"),
-    (test_avatar_creation_limit, "avatar creation limit"),
-    (test_avatar_deletion, "avatar deletion"),
-    (test_abort_account_deletion, "abort account deletion"),
-    (test_account_deletion, "account deletion"),
-    (test_admin_setup, "admin account setup"),
-    (test_preferences, "set and show preferences"),
-    (test_crlf_eol, "send crlf from the client as eol"),
-    (test_telnet_iac, "escape stray telnet iac bytes"),
-    (test_telnet_unknown_command, "strip unknown telnet command"),
-    (test_telnet_unknown_option, "log unknown telnet option"),
-    (test_admin_restriction, "restricted admin commands"),
-    (test_admin_help, "admin help"),
-    (test_help, "help command"),
-    (test_abbrev, "command abbreviation"),
-    (test_reload, "reload"),
-    (test_set_facet, "set facet"),
-    (test_set_refused, "refuse altering read-only element"),
-    (test_show_version, "show version and diagnostic info"),
-    (test_show_time, "show elapsed world clock increments"),
-    (test_show_files, "show a list of loaded files"),
-    (test_show_file, "show nodes from a specific file"),
-    (test_show_groups, "show groups"),
-    (test_show_group, "show group"),
-    (test_show_element, "show element"),
-    (test_show_result, "show result of a python expression"),
-    (test_show_log, "show log"),
-    (test_custom_loglevel, "custom loglevel"),
-    (test_invalid_loglevel, "invalid loglevel"),
-    (test_log_no_errors, "no errors logged"),
-    (final_cleanup, "delete remaining accounts"),
+dialogue = {
+    test_account0_setup: "first account setup",
+    test_account1_setup: "second account setup",
+    test_actor_appears: "actor spontaneous appearance",
+    test_explicit_punctuation: "explicit punctuation",
+    test_implicit_punctuation: "implicit punctuation",
+    test_typo_replacement: "typo replacement",
+    test_sentence_capitalization: "sentence capitalization",
+    test_chat_mode: "chat mode",
+    test_wrapping: "wrapping",
+    test_forbid_ansi_input: "raw escape input is filtered",
+    test_escape_macros: "replacement macros are escaped",
+    test_movement: "movement",
+    test_actor_disappears: "actor spontaneous disappearance",
+    test_abort_avatar_deletion: "abort avatar deletion",
+    test_avatar_creation_limit: "avatar creation limit",
+    test_avatar_deletion: "avatar deletion",
+    test_abort_account_deletion: "abort account deletion",
+    test_account_deletion: "account deletion",
+    test_admin_setup: "admin account setup",
+    test_preferences: "set and show preferences",
+    test_crlf_eol: "send crlf from the client as eol",
+    test_telnet_iac: "escape stray telnet iac bytes",
+    test_telnet_unknown_command: "strip unknown telnet command",
+    test_telnet_unknown_option: "log unknown telnet option",
+    test_admin_restriction: "restricted admin commands",
+    test_admin_help: "admin help",
+    test_help: "help command",
+    test_abbrev: "command abbreviation",
+    test_reload: "reload",
+    test_set_facet: "set facet",
+    test_set_refused: "refuse altering read-only element",
+    test_show_version: "show version and diagnostic info",
+    test_show_time: "show elapsed world clock increments",
+    test_show_files: "show a list of loaded files",
+    test_show_file: "show nodes from a specific file",
+    test_show_groups: "show groups",
+    test_show_group: "show group",
+    test_show_element: "show element",
+    test_evaluate: "show results of python expressions",
+    test_debug_restricted: "only admins can run debug commands",
+    test_debug_disabled: "debugging commands only in debug mode",
+    test_show_log: "show log",
+    test_custom_loglevel: "custom loglevel",
+    test_invalid_loglevel: "invalid loglevel",
+    test_log_no_errors: "no errors logged",
+    final_cleanup: "delete remaining accounts",
+}
+
+debug_tests = (
+    test_evaluate,
+)
+
+nondebug_tests = (
+    test_debug_disabled,
 )
 
 
 def start_service(config):
     # Clean up any previously run daemon which didn't terminate
     if os.path.exists(pidfile):
-        pidfd = open(pidfile)
-        pid = int(pidfd.read())
+        with open(pidfile) as pidfd:
+            pid = int(pidfd.read())
         try:
             # Stop the running service
             os.kill(pid, 15)
-            time.sleep(1)
         except ProcessLookupError:
             # If there was no process, just remove the stale PID file
             os.remove(pidfile)
@@ -443,9 +469,7 @@ def start_service(config):
 
     # Clean up any previous test output
     for f in pathlib.Path(".").glob("capture_*.log"):
-        # have to use .name here since remove() doesn't support passing a
-        # PosixPath argument until Python3.6
-        os.remove(f.name)
+        os.remove(f)
     for d in ("data", "var"):
         shutil.rmtree(d, ignore_errors=True)
 
@@ -453,8 +477,7 @@ def start_service(config):
     service = subprocess.Popen(("mudpy", config),
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
-    time.sleep(1)
-    return(service)
+    return service
 
 
 def stop_service(service):
@@ -462,7 +485,7 @@ def stop_service(service):
 
     # The no-op case when no service was started
     if service is None:
-        return(success)
+        return success
 
     # This handles when the service is running as a direct child process
     service.terminate()
@@ -473,8 +496,8 @@ def stop_service(service):
 
     # This cleans up a daemonized and disassociated service
     if os.path.exists(pidfile):
-        pidfd = open(pidfile)
-        pid = int(pidfd.read())
+        with open(pidfile) as pidfd:
+            pid = int(pidfd.read())
         try:
             # Stop the running service
             os.kill(pid, 15)
@@ -490,13 +513,13 @@ def stop_service(service):
     # Log the contents of stdout and stderr, if any
     stdout, stderr = service.communicate()
     tlog("\nRecording stdout as capture_stdout.log.")
-    serviceout = open("capture_stdout.log", "w")
-    serviceout.write(stdout.decode("utf-8"))
+    with open("capture_stdout.log", "w") as serviceout:
+        serviceout.write(stdout.decode("utf-8"))
     tlog("\nRecording stderr as capture_stderr.log.")
-    serviceerr = open("capture_stderr.log", "w")
-    serviceerr.write(stderr.decode("utf-8"))
+    with open("capture_stderr.log", "w") as serviceerr:
+        serviceerr.write(stderr.decode("utf-8"))
 
-    return(success)
+    return success
 
 
 def tlog(message, quiet=False):
@@ -519,6 +542,39 @@ def option_callback(telnet_socket, command, option):
         telnet_socket.send(telnetlib.IAC + telnetlib.DONT + option)
 
 
+def check_debug():
+    if len(sys.argv) > 1:
+        with open(sys.argv[1]) as config_fd:
+            config = yaml.safe_load(config_fd)
+        return config.get(".mudpy.limit.debug", False)
+    return False
+
+
+def connect_client(luser, service):
+    # Try multiple times to connect, with an exponential backoff
+    for retry in range(5):
+        try:
+            # Skipping the retry=0 case gives an immediate first attempt
+            if retry:
+                time.sleep((2 ** retry) / 10)
+            luser.open("::1", 4000)
+            # Attempt to poll the connection, closing if unusable
+            try:
+                luser.fill_rawq()
+            except ConnectionResetError:
+                luser.close()
+                continue
+            # Short-circuit if we get this far, connection is safe to use
+            return luser
+        except ConnectionRefusedError:
+            continue
+    else:
+        # Connection retries have been exhausted, so give up
+        tlog("\nERROR: Client could not connect.\n")
+        stop_service(service)
+        sys.exit(1)
+
+
 def main():
     captures = ["", "", ""]
     lusers = [telnetlib.Telnet(), telnetlib.Telnet(), telnetlib.Telnet()]
@@ -532,9 +588,16 @@ def main():
             tlog("\nERROR: Service did not start.\n")
             sys.exit(1)
     for luser in lusers:
-        luser.open("::1", 4000)
+        connect_client(luser, service)
         luser.set_option_negotiation_callback(option_callback)
-    for test, description in dialogue:
+    selected_dialogue = dict(dialogue)
+    if check_debug():
+        for test in nondebug_tests:
+            del selected_dialogue[test]
+    else:
+        for test in debug_tests:
+            del selected_dialogue[test]
+    for test, description in selected_dialogue.items():
         tlog("\nTesting %s..." % description)
         test_start = time.time()
         for conversant, question, answer in test:
@@ -544,11 +607,7 @@ def main():
                 index, match, received = lusers[conversant].expect(
                     [re.compile(question.encode("utf-8"), flags=re.DOTALL)], 5)
                 captures[conversant] += received.decode("utf-8")
-            except ConnectionResetError:
-                tlog("\nERROR: Unable to connect to server.")
-                success = False
-                break
-            except EOFError:
+            except (ConnectionResetError, EOFError):
                 tlog("\nERROR: luser%s premature disconnection expecting:\n\n"
                      "%s\n\n"
                      "Check the end of capture_%s.log for received data."
